package com.alterdekim.game.service;

import com.alterdekim.game.component.game.*;
import com.alterdekim.game.component.game.avatar.*;
import com.alterdekim.game.component.game.inventory.Inventory;
import com.alterdekim.game.component.game.inventory.InventoryList;
import com.alterdekim.game.component.game.inventory.InventoryTab;
import com.alterdekim.game.component.game.response.init.UserAccount;
import com.alterdekim.game.component.game.response.init.UserInfo;
import com.alterdekim.game.component.game.response.init.UserInitInfo;
import com.alterdekim.game.component.game.response.init.UserPhone;
import com.alterdekim.game.entity.PhoneSkin;
import com.alterdekim.game.entity.Role;
import com.alterdekim.game.entity.User;
import com.alterdekim.game.entity.UserProperty;
import com.alterdekim.game.repository.RoleRepository;
import com.alterdekim.game.repository.UserPropertyRepository;
import com.alterdekim.game.repository.UserRepository;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.crypto.password.PasswordEncoder;
import org.springframework.stereotype.Service;

import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

@Service
public class UserService {

    @Autowired
    private UserRepository userRepository;
    @Autowired
    private RoleRepository roleRepository;
    @Autowired
    private PasswordEncoder passwordEncoder;
    @Autowired
    private UserPropertyRepository propertyRepository;

    @Autowired
    private OnlineStatusService onlineStatusService;

    @Autowired
    private AvatarInventoryService inventoryService;

    @Autowired
    private PhoneIconService phoneIconService;

    public User findByUsername(String username) {
        return userRepository.findByUsername(username);
    }

    public long saveUser(String username, String password) {
        User user = new User();
        user.setUsername(username);
        user.setPassword(passwordEncoder.encode(password));
        Role role = roleRepository.findByName("ROLE_USER");
        if(role == null){
            role = checkRoleExist();
        }
        user.setRoles(Collections.singletonList(role));
        return userRepository.save(user).getId();
    }

    private Role checkRoleExist() {
        Role role = new Role();
        role.setName("ROLE_USER");
        return roleRepository.save(role);
    }

    public Optional<Object> findUserProperty(Long userId, PlayerProperties type) {
        Optional<UserProperty> res = this.propertyRepository.findByUserIdAndTag(userId, type);
        if( res.isEmpty() ) return Optional.empty();
        UserProperty property = res.get();
        return Optional.of(switch (property.getType().getValueType()) {
            case String -> property.getStringVal();
            case Integer -> property.getIntegerVal();
            case Boolean -> property.getBooleanVal();
            case Date -> property.getDateVal();
            case Enum -> property.getType() == PlayerProperties.RoleFlags ? RoleFlags.valueOf(property.getStringVal()) : ClubAccessType.valueOf(property.getStringVal());
        });
    }

    public void pushUserProperty(Long userId, PlayerProperties type, Object val) {
        this.propertyRepository.deleteByUserIdAndType(userId, type);
        this.propertyRepository.save(switch (type.getValueType()) {
            case String -> new UserProperty(userRepository.getReferenceById(userId), (String) val, type);
            case Integer -> new UserProperty(userRepository.getReferenceById(userId), (Integer) val, type);
            case Boolean -> new UserProperty(userRepository.getReferenceById(userId), (Boolean) val, type);
            case Date -> new UserProperty(userRepository.getReferenceById(userId), (LocalDateTime) val, type);
            case Enum -> new UserProperty(userRepository.getReferenceById(userId), String.valueOf(val), type);
        });
    }

    public RoleFlags getRoleFlags(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.RoleFlags);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.RoleFlags, RoleFlags.SA);
            return RoleFlags.SA;
        }
        return (RoleFlags) res.get();
    }

    public Boolean getIsClubPresent(Long userId) {
        // todo: Implement clubs;
        return false;
    }

    public ClubAccessType getClubAccessType(Long userId) {
        Optional<Object> res = this.findUserProperty(userId, PlayerProperties.ClubAccessType);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, PlayerProperties.ClubAccessType, ClubAccessType.CLOSED);
            return ClubAccessType.CLOSED;
        }
        return (ClubAccessType) res.get();
    }

    public Boolean getIsFriend(Long userId) {
        // todo: Implement friends;
        return false;
    }

    public void setWeaponsCount(Long userId, Integer count) {
        this.pushUserProperty(userId, PlayerProperties.WeaponsCount, count);
    }
    
    public Integer getIntegerUserProperty(Long userId, PlayerProperties prop, Integer def) {
        Optional<Object> res = this.findUserProperty(userId, prop);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, prop, def);
            return def;
        }
        return res.map(o -> (Integer) o).get();
    }

    public Boolean getBooleanUserProperty(Long userId, PlayerProperties prop, Boolean def) {
        Optional<Object> res = this.findUserProperty(userId, prop);
        if( res.isEmpty() ) {
            this.pushUserProperty(userId, prop, def);
            return def;
        }
        return res.map(o -> (Boolean) o).get();
    }

    public Optional<String> getUsernameById(Long userId) {
        return this.userRepository.findById(userId).map(User::getUsername);
    }

    public UserAccount getUserAccountByUserId(Long userId) {
        return new UserAccount(
                getIntegerUserProperty(userId, PlayerProperties.PhoneCardBalance, 10),
                getIntegerUserProperty(userId, PlayerProperties.WeaponsCount, 15),
                getIntegerUserProperty(userId, PlayerProperties.OxygenUnits, 1)
        );
    }

    public void setDefaultPhone(Long userId, Long phoneId) {
        this.inventoryService.setDefault(userId, phoneId, AvatarInventoryType.Phones);
    }

    public void setDefaultBackground(Long userId, Long bgId) {
        this.inventoryService.setDefault(userId, bgId, AvatarInventoryType.Backgrounds);
    }

    public UserPhone getUserPhoneByUserId(Long userId) {
        Long phoneId = inventoryService.getSelectedPhoneId(userId);
        return new UserPhone(
                999,
                phoneIconService.findSkinById(phoneId).map(PhoneSkin::getSkinMRId).orElse(401L),
                phoneId.intValue(),
                getIntegerUserProperty(userId, PlayerProperties.PhoneCardBalance, 10)
        );
    }

    public Inventory getUserAvatarByUserId(Long userId) {
        return new Inventory(
            new InventoryTab(199, List.of(
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Body),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Legs)
            )),
            new InventoryTab(205, List.of(
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Horn),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Beak),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Mouth),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Nose),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Eyes),
                    inventoryService.getBodyPartInventoryGroup(userId, BodyPartType.Ears)
            )),
            new InventoryTab(206, List.of(
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Clothes),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Caps),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Glasses),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Boots),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Bangles),
                    inventoryService.getClothInventoryGroup(userId, GoodClothType.Suits)
            )),
            new InventoryList(230, inventoryService.getBackgroundsByUserId(userId)),
            new InventoryList(231, inventoryService.getPhonesByUserId(userId)),
            new InventoryList(211, Collections.emptyList())
        );
    }

    public UserInfo getUserInfoByUserId(Long userId, boolean isForUserInit) {
        // todo: implement hardcoded stuff
        return new UserInfo(
                userId,
                !isForUserInit ? userId : null,
                getRoleFlags(userId),
                !isForUserInit ? true : null,
                getBooleanUserProperty(userId, PlayerProperties.IsLimited, false),
                onlineStatusService.isOnline(userId),
                getBooleanUserProperty(userId, PlayerProperties.IsMinorSecretAgent, false),
                getBooleanUserProperty(userId, PlayerProperties.IsHomeLocked, true),
                getBooleanUserProperty(userId, PlayerProperties.IsClubLocked, true),
                getIsClubPresent(userId),
                getClubAccessType(userId),
                null,
                getIsFriend(userId),
                getIntegerUserProperty(userId, PlayerProperties.Experience, 0),
                getIntegerUserProperty(userId, PlayerProperties.Level, 1),
                getIntegerUserProperty(userId, PlayerProperties.MagicLevel, 1),
                getIntegerUserProperty(userId, PlayerProperties.RaceLevel, 1),
                inventoryService.getSelectedPhoneId(userId).intValue(),
                inventoryService.getUsedBackground(userId),
                getIntegerUserProperty(userId, PlayerProperties.CampId, 1),
                getIntegerUserProperty(userId, PlayerProperties.NiftTotalCount, 1),
                "Server",
                LocalDateTime.now(),
                getIntegerUserProperty(userId, PlayerProperties.InstructorLevel, 1),
                getIntegerUserProperty(userId, PlayerProperties.UsualTickets, 1000),
                getIntegerUserProperty(userId, PlayerProperties.MagicTickets, 500),
                getIntegerUserProperty(userId, PlayerProperties.VisaId, 10),
                getIntegerUserProperty(userId, PlayerProperties.CurrentUseCount, 10),
                getIntegerUserProperty(userId, PlayerProperties.LastNewspaperId, 10),
                getBooleanUserProperty(userId, PlayerProperties.HasSnFriends, true),
                getIntegerUserProperty(userId, PlayerProperties.MinsLeft, 10),
                getIntegerUserProperty(userId, PlayerProperties.TimesUsed, 10),
                this.getUsernameById(userId).get(),
                inventoryService.getAchievementsByUserId(userId),
                isForUserInit ? null : inventoryService.getMagicAbilitiesItemsByUserId(userId)
        );
    }

    public List<IBodyAvatarItem> getOtherUserAvatar(Long userId) {
        return this.inventoryService.getWholeAvatarByUserId(userId);
    }

    public UserInitInfo getUserInitInfoByUserId(Long userId) {
        return new UserInitInfo(
                getUserAccountByUserId(userId),
                getUserInfoByUserId(userId, true),
                getUserPhoneByUserId(userId),
                inventoryService.getMagicAbilitiesItemsByUserId(userId),
                inventoryService.getSmilesByUserId(userId)
        );
    }

    public PlayerAvatar getAvatarById(long userId, Player p) {
        Map<String, BodyAvatarItem> body = new HashMap<>();
        List<BodyAvatarItem> i = inventoryService.getWholeAvatarByUserId(userId).stream().map(BodyAvatarItem::fromInterface).collect(Collectors.toList());
        IntStream.range(0, i.size()).boxed().forEach(g -> body.put(String.valueOf(g+1), i.get(g)));
        UserAvatar a = new UserAvatar(getIntegerUserProperty(userId, PlayerProperties.Level, 1), getUsernameById(userId).get(), p.getX(), p.getY(), p.getState(), body);
        return new PlayerAvatar(String.valueOf(getRoleFlags(userId).getValue()), a);
    }

    public void removeUser(long userId) {
        this.userRepository.deleteById(userId);
    }

    public Long getUsersCount() {
        return this.userRepository.findUsersCount();
    }
}
